@@ -30,6 +30,7 @@ from manila.share import configuration
 from manila.share import driver
 from manila.share.drivers.netapp.dataontap.client import api as netapp_api
 from manila.share.drivers.netapp.dataontap.client import client_cmode
+from manila.share.drivers.netapp.dataontap.client import client_cmode_rest
 from manila.share.drivers.netapp import options as na_opts
 from manila.share.drivers.netapp import utils as na_utils
 from manila.share import utils as share_utils
@@ -66,21 +67,53 @@ def get_backend_configuration(backend_name):
     config.append_config_values(na_opts.netapp_support_opts)
     config.append_config_values(na_opts.netapp_provisioning_opts)
     config.append_config_values(na_opts.netapp_data_motion_opts)
+    config.append_config_values(na_opts.netapp_proxy_opts)
+    config.append_config_values(na_opts.netapp_backup_opts)
 
     return config
 
 
+def get_backup_configuration(backup_type):
+    config_stanzas = CONF.list_all_sections()
+    if backup_type not in config_stanzas:
+        msg = _("Could not find backup_type stanza %(backup_type)s in "
+                "configuration which is required for backup workflows "
+                "with the source share. Available stanzas are "
+                "%(stanzas)s")
+        params = {
+            "stanzas": config_stanzas,
+            "backup_type": backup_type,
+        }
+        raise exception.BadConfigurationException(reason=msg % params)
+    config = configuration.Configuration(driver.share_opts,
+                                         config_group=backup_type)
+    config.append_config_values(na_opts.netapp_backup_opts)
+    return config
+
+
 def get_client_for_backend(backend_name, vserver_name=None):
     config = get_backend_configuration(backend_name)
-    client = client_cmode.NetAppCmodeClient(
-        transport_type=config.netapp_transport_type,
-        ssl_cert_path=config.netapp_ssl_cert_path,
-        username=config.netapp_login,
-        password=config.netapp_password,
-        hostname=config.netapp_server_hostname,
-        port=config.netapp_server_port,
-        vserver=vserver_name or config.netapp_vserver,
-        trace=na_utils.TRACE_API)
+    if config.netapp_use_legacy_client:
+        client = client_cmode.NetAppCmodeClient(
+            transport_type=config.netapp_transport_type,
+            ssl_cert_path=config.netapp_ssl_cert_path,
+            username=config.netapp_login,
+            password=config.netapp_password,
+            hostname=config.netapp_server_hostname,
+            port=config.netapp_server_port,
+            vserver=vserver_name or config.netapp_vserver,
+            trace=na_utils.TRACE_API)
+    else:
+        client = client_cmode_rest.NetAppRestClient(
+            transport_type=config.netapp_transport_type,
+            ssl_cert_path=config.netapp_ssl_cert_path,
+            username=config.netapp_login,
+            password=config.netapp_password,
+            hostname=config.netapp_server_hostname,
+            port=config.netapp_server_port,
+            vserver=vserver_name or config.netapp_vserver,
+            async_rest_timeout=config.netapp_rest_operation_timeout,
+            trace=na_utils.TRACE_API)
 
     return client
 
@@ -162,6 +195,7 @@ class DataMotionSession(object):
             source_volume=src_volume_name, dest_volume=dest_volume_name,
             desired_attributes=['relationship-status',
                                 'mirror-state',
+                                'schedule',
                                 'source-vserver',
                                 'source-volume',
                                 'last-transfer-end-timestamp',
@@ -186,13 +220,14 @@ class DataMotionSession(object):
             source_share_obj)
 
         # 1. Create SnapMirror relationship
-        # TODO(ameade): Change the schedule from hourly to a config value
+        config = get_backend_configuration(dest_backend)
+        schedule = config.netapp_snapmirror_schedule
         dest_client.create_snapmirror_vol(src_vserver,
                                           src_volume_name,
                                           dest_vserver,
                                           dest_volume_name,
                                           relationship_type,
-                                          schedule='hourly')
+                                          schedule=schedule)
 
         # 2. Initialize async transfer of the initial data
         dest_client.initialize_snapmirror_vol(src_vserver,
@@ -305,7 +340,8 @@ class DataMotionSession(object):
                 source_vserver=source_vserver, dest_vserver=dest_vserver,
                 desired_attributes=['relationship-status', 'mirror-state']
             )[0]
-            if snapmirror.get('relationship-status') != 'quiesced':
+            if snapmirror.get('relationship-status') not in ['quiesced',
+                                                             'paused']:
                 raise exception.ReplicationException(
                     reason="Snapmirror relationship is not quiesced.")
 
@@ -316,7 +352,8 @@ class DataMotionSession(object):
                                              dest_vserver,
                                              clear_checkpoint=False)
 
-    def quiesce_then_abort(self, source_share_obj, dest_share_obj):
+    def quiesce_then_abort(self, source_share_obj, dest_share_obj,
+                           quiesce_wait_time=None):
         dest_volume, dest_vserver, dest_backend = (
             self.get_backend_info_for_share(dest_share_obj))
         dest_client = get_client_for_backend(dest_backend,
@@ -332,7 +369,9 @@ class DataMotionSession(object):
                                            dest_volume)
 
         config = get_backend_configuration(dest_backend)
-        retries = config.netapp_snapmirror_quiesce_timeout / 5
+        timeout = (
+            quiesce_wait_time or config.netapp_snapmirror_quiesce_timeout)
+        retries = int(timeout / 5) or 1
 
         @utils.retry(retry_param=exception.ReplicationException,
                      interval=5,
@@ -344,7 +383,8 @@ class DataMotionSession(object):
                 source_volume=src_volume, dest_volume=dest_volume,
                 desired_attributes=['relationship-status', 'mirror-state']
             )[0]
-            if snapmirror.get('relationship-status') != 'quiesced':
+            if snapmirror.get('relationship-status') not in ['quiesced',
+                                                             'paused']:
                 raise exception.ReplicationException(
                     reason="Snapmirror relationship is not quiesced.")
 
@@ -357,7 +397,8 @@ class DataMotionSession(object):
                                              dest_volume,
                                              clear_checkpoint=False)
 
-    def break_snapmirror(self, source_share_obj, dest_share_obj, mount=True):
+    def break_snapmirror(self, source_share_obj, dest_share_obj, mount=True,
+                         quiesce_wait_time=None):
         """Breaks SnapMirror relationship.
 
         1. Quiesce any ongoing snapmirror transfers
@@ -374,7 +415,8 @@ class DataMotionSession(object):
             source_share_obj)
 
         # 1. Attempt to quiesce, then abort
-        self.quiesce_then_abort(source_share_obj, dest_share_obj)
+        self.quiesce_then_abort(source_share_obj, dest_share_obj,
+                                quiesce_wait_time=quiesce_wait_time)
 
         # 2. Break SnapMirror
         dest_client.break_snapmirror_vol(src_vserver,
@@ -401,6 +443,27 @@ class DataMotionSession(object):
                                           dest_vserver,
                                           dest_volume_name)
 
+    def modify_snapmirror(self, source_share_obj, dest_share_obj,
+                          schedule=None):
+        """Modify SnapMirror relationship: set schedule"""
+        dest_volume_name, dest_vserver, dest_backend = (
+            self.get_backend_info_for_share(dest_share_obj))
+        dest_client = get_client_for_backend(dest_backend,
+                                             vserver_name=dest_vserver)
+
+        src_volume_name, src_vserver, __ = self.get_backend_info_for_share(
+            source_share_obj)
+
+        if schedule is None:
+            config = get_backend_configuration(dest_backend)
+            schedule = config.netapp_snapmirror_schedule
+
+        dest_client.modify_snapmirror_vol(src_vserver,
+                                          src_volume_name,
+                                          dest_vserver,
+                                          dest_volume_name,
+                                          schedule=schedule)
+
     def resume_snapmirror(self, source_share_obj, dest_share_obj):
         """Resume SnapMirror relationship from a quiesced state."""
         dest_volume_name, dest_vserver, dest_backend = (
@@ -475,14 +538,14 @@ class DataMotionSession(object):
                                                        replica_vserver)
 
         # 3. create
-        # TODO(ameade): Update the schedule if needed.
         relationship_type = na_utils.get_relationship_type(is_flexgroup)
+        schedule = replica_config.netapp_snapmirror_schedule
         replica_client.create_snapmirror_vol(new_src_vserver,
                                              new_src_volume_name,
                                              replica_vserver,
                                              replica_volume_name,
                                              relationship_type,
-                                             schedule='hourly')
+                                             schedule=schedule)
 
         # 4. resync
         replica_client.resync_snapmirror_vol(new_src_vserver,
@@ -840,3 +903,88 @@ class DataMotionSession(object):
             msg = _("Unable to release the snapmirror from source volume %s. "
                     "Retries exhausted. Aborting") % src_volume_name
             raise exception.NetAppException(message=msg)
+
+    def cleanup_previous_snapmirror_relationships(self, replica, replica_list):
+        """Cleanup previous snapmirrors relationships for replica."""
+        LOG.debug("Cleaning up old snapmirror relationships for replica %s.",
+                  replica['id'])
+        src_vol_name, src_vserver, src_backend = (
+            self.get_backend_info_for_share(replica))
+        src_client = get_client_for_backend(src_backend,
+                                            vserver_name=src_vserver)
+
+        # replica_list may contain the replica we are trying to clean up
+        destinations = (r for r in replica_list if r['id'] != replica['id'])
+
+        for destination in destinations:
+            dest_vol_name, dest_vserver, _ = (
+                self.get_backend_info_for_share(destination))
+            try:
+                src_client.release_snapmirror_vol(
+                    src_vserver, src_vol_name, dest_vserver, dest_vol_name)
+            except netapp_api.NaApiError as e:
+                if (e.code == netapp_api.EOBJECTNOTFOUND or
+                        e.code == netapp_api.ESOURCE_IS_DIFFERENT or
+                        "(entry doesn't exist)" in e.message):
+                    LOG.debug(
+                        'Snapmirror destination %s no longer exists for '
+                        'replica %s.', destination['id'], replica['id'])
+                else:
+                    LOG.exception(
+                        'Error releasing snapmirror destination %s for '
+                        'replica %s.', destination['id'], replica['id'])
+
+    def get_most_available_aggr_of_vserver(self, vserver_client):
+        """Get most available aggregate"""
+        aggrs_space_attr = vserver_client.get_vserver_aggregate_capacities()
+        if not aggrs_space_attr:
+            return None
+        aggr_list = list(aggrs_space_attr.keys())
+        most_available_aggr = aggr_list[0]
+        for aggr in aggr_list:
+            if (aggrs_space_attr.get(aggr).get('available')
+                    > aggrs_space_attr.get(
+                        most_available_aggr).get('available')):
+                most_available_aggr = aggr
+        return most_available_aggr
+
+    def initialize_and_wait_snapmirror_vol(self, vserver_client,
+                                           source_vserver, source_volume,
+                                           dest_vserver, dest_volume,
+                                           source_snapshot=None,
+                                           transfer_priority=None,
+                                           timeout=300):
+        """Initialize and wait for SnapMirror relationship"""
+        interval = 10
+        retries = (timeout / interval or 1)
+        vserver_client.initialize_snapmirror_vol(
+            source_vserver,
+            source_volume,
+            dest_vserver,
+            dest_volume,
+            source_snapshot=source_snapshot,
+            transfer_priority=transfer_priority,
+        )
+
+        @utils.retry(exception.NetAppException, interval=interval,
+                     retries=retries, backoff_rate=1)
+        def wait_for_initialization():
+            source_path = f"{source_vserver}:{source_volume}"
+            des_path = f"{dest_vserver}:{dest_volume}"
+            snapmirror_info = vserver_client.get_snapmirrors(
+                source_path=source_path, dest_path=des_path)
+            relationship_status = snapmirror_info[0].get("relationship-status")
+            if relationship_status == "idle":
+                return
+            else:
+                msg = (_('Snapmirror relationship status is: %s. Waiting '
+                         'until it has been initialized.') %
+                       relationship_status)
+                raise exception.NetAppException(message=msg)
+
+        try:
+            wait_for_initialization()
+        except exception.NetAppException:
+            msg = _("Timed out while wait for SnapMirror relationship to "
+                    "be initialized")
+            raise exception.NetAppException(message=msg)
